from __future__ import print_function
from cbapi.six import PY3
import sys
import time
import argparse
import cmd
import codecs
import logging
import ntpath
import shutil
import subprocess
from optparse import OptionParser
from collections import defaultdict
import validators
import hashlib

from cbapi.protection import CbEnterpriseProtectionAPI
from cbapi.psc import CbPSCBaseAPI
from cbapi.psc.defense import CbDefenseAPI
from cbapi.psc.threathunter import CbThreatHunterAPI
from cbapi.psc.livequery import CbLiveQueryAPI
from cbapi.response import CbEnterpriseResponseAPI

log = logging.getLogger(__name__)


# Example scripts: we want to make sure that sys.stdout is using utf-8 encoding. See issue #36.
if not PY3:
    sys.stdout = codecs.getwriter('utf8')(sys.stdout)


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def build_cli_parser(description="Cb Example Script"):
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("--cburl", help="CB server's URL.  e.g., http://127.0.0.1 ")
    parser.add_argument("--apitoken", help="API Token for Carbon Black server")
    parser.add_argument("--orgkey", help="Organization key value for Carbon Black server")
    parser.add_argument("--no-ssl-verify", help="Do not verify server SSL certificate.", action="store_true",
                        default=False)
    parser.add_argument("--profile", help="profile to connect", default="default")
    parser.add_argument("--verbose", help="enable debug logging", default=False, action='store_true')

    return parser


def disable_insecure_warnings():
    import requests.packages.urllib3
    requests.packages.urllib3.disable_warnings()


def get_cb_response_object(args):
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("cbapi").setLevel(logging.DEBUG)
        logging.getLogger("__main__").setLevel(logging.DEBUG)

    if args.cburl and args.apitoken:
        cb = CbEnterpriseResponseAPI(url=args.cburl, token=args.apitoken, ssl_verify=(not args.no_ssl_verify))
    else:
        cb = CbEnterpriseResponseAPI(profile=args.profile)

    return cb


def get_cb_protection_object(args):
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("cbapi").setLevel(logging.DEBUG)
        logging.getLogger("__main__").setLevel(logging.DEBUG)

    if args.cburl and args.apitoken:
        cb = CbEnterpriseProtectionAPI(args.cburl, args.apitoken)
    else:
        cb = CbEnterpriseProtectionAPI(profile=args.profile)

    return cb


def get_cb_psc_object(args):
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("cbapi").setLevel(logging.DEBUG)
        logging.getLogger("__main__").setLevel(logging.DEBUG)

    if args.cburl and args.apitoken:
        cb = CbPSCBaseAPI(url=args.cburl, token=args.apitoken, ssl_verify=(not args.no_ssl_verify))
    else:
        cb = CbPSCBaseAPI(profile=args.profile)

    return cb


def get_cb_defense_object(args):
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("cbapi").setLevel(logging.DEBUG)
        logging.getLogger("__main__").setLevel(logging.DEBUG)

    if args.cburl and args.apitoken:
        cb = CbDefenseAPI(url=args.cburl, token=args.apitoken, ssl_verify=(not args.no_ssl_verify))
    else:
        cb = CbDefenseAPI(profile=args.profile)

    return cb


def get_cb_threathunter_object(args):
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("cbapi").setLevel(logging.DEBUG)
        logging.getLogger("__main__").setLevel(logging.DEBUG)

    if args.cburl and args.apitoken:
        cb = CbThreatHunterAPI(url=args.cburl, token=args.apitoken, ssl_verify=(not args.no_ssl_verify))
    else:
        cb = CbThreatHunterAPI(profile=args.profile)

    return cb


def get_cb_livequery_object(args):
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("cbapi").setLevel(logging.DEBUG)
        logging.getLogger("__main__").setLevel(logging.DEBUG)

    if args.cburl and args.apitoken and args.orgkey:
        cb = CbLiveQueryAPI(url=args.cburl, token=args.apitoken, org_key=args.orgkey,
                            ssl_verify=(not args.no_ssl_verify))
    else:
        cb = CbLiveQueryAPI(profile=args.profile)

    return cb


def get_object_by_name_or_id(cb, cls, name_field="name", id=None, name=None, force_init=True):
    clsname = cls.__name__
    try:
        if id:
            attempted_to_find = "ID of {0:d}".format(id)
            objs = [cb.select(cls, id, force_init=force_init)]
        else:
            attempted_to_find = "name {0:s}".format(name)
            objs = cb.select(cls).where("{0}:{1}".format(name_field, name))[::]
            if not len(objs):
                raise Exception("No {0}s match".format(clsname))
    except Exception as e:
        raise Exception("Could not find {0} with {1:s}: {2:s}".format(clsname, attempted_to_find, str(e)))
    else:
        return objs


def read_iocs(cb, file=sys.stdin):
    iocs = defaultdict(list)
    report_id = hashlib.md5()
    report_id.update(str(time.time()).encode("utf-8"))

    for idx, line in enumerate(sys.stdin):
        line = line.rstrip("\r\n")
        report_id.update(line.encode("utf-8"))
        if validators.md5(line):
            iocs["md5"].append(line)
        elif validators.sha256(line):
            eprint("line {}: sha256 provided but not yet supported by backend".format(idx + 1))
            iocs["sha256"].append(line)
        elif validators.ipv4(line):
            iocs["ipv4"].append(line)
        elif validators.ipv6(line):
            iocs["ipv6"].append(line)
        elif validators.domain(line):
            iocs["dns"].append(line)
        else:
            if cb.validate_query(line):
                query_ioc = {"search_query": line}
                iocs["query"].append(query_ioc)
            else:
                eprint("line {}: invalid query".format(idx + 1))

    return (report_id.hexdigest(), dict(iocs))

#
# Live Response
#


class QuitException(Exception):
    pass


class CliArgsException(Exception):
    pass


class CliHelpException(Exception):
    pass


class CliAttachError(Exception):
    pass


def split_cli(line):
    '''
    we'd like to use shlex.split() but that doesn't work well for
    windows types of things.  for now we'll take the easy approach
    and just split on space. We'll then cycle through and look for leading
    quotes and join those lines
    '''

    parts = line.split(' ')
    final = []

    while len(parts) > 0:

        tok = parts.pop(0)
        if (tok[:1] == '"'):
            tok = tok[1:]
            next = parts.pop(0)
            while(next[-1:] != '"' and len(parts) > 0):
                tok += ' ' + next
                next = parts.pop(0)

            if (next[-1:] == '"'):
                tok += ' ' + next[:-1]

        final.append(tok)
    return final


class CliArgs (OptionParser):
    def __init__(self, usage=''):
        OptionParser.__init__(self, add_help_option=False, usage=usage)

        self.add_option('-h', '--help', action='store_true', help='Display this help message.')

    def parse_line(self, line):
        args = split_cli(line)
        return self.parse_args(args)

    def parse_args(self, args, values=None):
        (opts, args) = OptionParser.parse_args(self, args=args, values=values)

        if (opts.help):
            self.print_help()
            raise CliHelpException()

        return (opts, args)

    def error(self, msg):
        raise CliArgsException(msg)


class CblrCli(cmd.Cmd):
    def __init__(self, cb, connect_callback):
        """
        Create a CbLR Command Line class

        :param cb: Connection to the Cb Enterprise Response API
        :param connect_callback: Callable to get a sensor object from the ``connect`` command
        :type cb: CbEnterpriseResponseAPI
        :return:
        """
        cmd.Cmd.__init__(self)

        # global variables
        # apply regardless of session state
        self.cb = cb
        self.connect_callback = connect_callback

        lr_session = None
        """:type lr_session: LiveResponseSession"""
        self.lr_session = lr_session

        self.reset()

    @property
    def prompt(self):
        if not self.lr_session:
            return "(unattached)> "
        else:
            return "{0}\\> ".format(self.cwd)

    def emptyline(self):
        pass

    def cmdloop(self, intro=None):
        while True:
            try:
                cmd.Cmd.cmdloop(self, intro)
            except CliHelpException:
                pass
            except QuitException:
                break
            except KeyboardInterrupt:
                break
            except CliAttachError:
                print("You must attach to a session")
                continue
            except CliArgsException as e:
                print("Error parsing arguments!\n %s" % e)
                continue
            except Exception as e:
                print("Error: %s" % e)
                continue

        if self.lr_session:
            self.lr_session.close()

    def _quit(self):
        raise QuitException("quit")

    def _is_path_absolute(self, path):

        if path.startswith('\\\\'):
            return True

        if (path[0].isalpha() and path[1:3] == ':\\'):
            return True

        return False

    def _is_path_drive_relative(self, path):

        if path == '\\':
            return True

        if path[0] == '\\' and path[1] != '\\':
            return True

        return False

    def _file_path_fixup(self, path):
        '''
        We have a pseudo-cwd that we use to
        base off all commands.  This means we
        need to figure out if a given path is relative,
        absolute, or file relative and calculate against
        the pseudo cwd.

        This function takes in a given file path arguemnt
        and performs the fixups.
        '''

        if (self._is_path_absolute(path)):
            return path
        elif (self._is_path_drive_relative(path)):
            return self.cwd[:2] + path
        else:
            return ntpath.join(self.cwd + '\\', path)

    def _stat(self, path):
        '''
        Look to see if a given path exists
        on the sensor and whether that is a
        file or directory.

        :param path: a sensor path
        :return: None, "dir", or "file"
        '''
        if path.endswith('\\'):
            path = path[:-1]

        ret = self.lr_session.list_directory(path)
        if not ret or not len(ret):
            return None

        if 'DIRECTORY' in ret[0]['attributes']:
            return "dir"
        else:
            return "file"

    ################################
    # pseudo commands and session commands
    #
    # they don't change state on the sensor
    # (except start a session)
    #####################

    def _needs_attached(self):
        if not self.lr_session:
            raise CliAttachError()

    def do_cd(self, line):
        '''
        Pseudo Command: cd

        Description:
        Change the psuedo current working directory

        Note: The shell keeps a pseudo working directory
        that allows the user to change directory without
        changing the directory of the working process on sensor.

        Args:
        cd <Directory>
        '''
        self._needs_attached()

        path = self._file_path_fixup(line)
        path = ntpath.abspath(path)
        type = self._stat(path)
        if (type != "dir"):
            print("Error: Path %s does not exist" % path)
            return
        else:
            self.cwd = path

        # cwd never has a trailing \
        if self.cwd[-1:] == '\\':
            self.cwd = self.cwd[:-1]

        log.info("Changed directory to {0}".format(self.cwd))

    def do_cat(self, line):
        self._needs_attached()
        gfile = self._file_path_fixup(line)
        shutil.copyfileobj(self.lr_session.get_raw_file(gfile), sys.stdout)

    def do_pwd(self, line):
        '''
        Pseudo Command: pwd

        Description:
        Print the pseudo current working directory

        Note: The shell keeps a pseudo working directory
        that allows the user to change directory without
        changing the directory of the working process on sensor.

        Args:
        pwd

        '''

        self._needs_attached()

        if (len(self.cwd) == 2):
            # it's c: - put a slash on the end
            print(self.cwd + '\\')
        else:
            print(self.cwd)
        print("")

    def do_connect(self, line):
        """
        Command: connect

        Description:
        Connect to a sensor given the sensor ID or the sensor hostname.

        Args:
        connect SENSOR_ID | SENSOR_HOSTNAME
        """
        if not line:
            raise CliArgsException("Need argument: sensor ID or hostname")

        sensor = self.connect_callback(self.cb, line)
        self.lr_session = sensor.lr_session()

        print("Session: {0}".format(self.lr_session.session_id))
        print("  Available Drives: %s" % ' '.join(self.lr_session.session_data.get('drives', [])))

        # look up supported commands
        print("  Supported Commands: %s" % ', '.join(self.lr_session.session_data.get('supported_commands', [])))
        print("  Working Directory: %s" % self.cwd)

        log.info("Attached to sensor {0}".format(sensor._model_unique_id))

    def do_detach(self, line):
        self.reset()

    def reset(self):
        self.cwd = "c:"

        if not self.lr_session:
            return

        self.lr_session.close()
        self.lr_session = None

    #############################
    # real commands
    #
    # these change state on the senesor
    ##############################

    def do_archive(self, line):
        filename = line
        shutil.copyfileobj(self.lr_session.get_session_archive(), open(filename, "wb+"))

    def do_ps(self, line):
        '''
        Command: ps

        Description:
        List the processes running on the sensor.

        Args:
        ps [OPTIONS]

        OPTIONS:
        -v  - Display verbose info about each process
        -p <Pid> - Display only the given pid
        '''
        self._needs_attached()

        p = CliArgs(usage='ps [OPTIONS]')
        p.add_option('-v', '--verbose', default=False, action='store_true',
                     help='Display verbose info about each process')
        p.add_option('-p', '--pid', default=None, help='Display only the given pid')
        (opts, args) = p.parse_line(line)

        if opts.pid:
            opts.pid = int(opts.pid)

        processes = self.lr_session.list_processes()

        for p in processes:
            if (opts.pid and p['pid'] == opts.pid) or opts.pid is None:
                if opts.verbose:
                    print("Process: %5d : %s" % (p['pid'], ntpath.basename(p['path'])))
                    print("  Guid:        %s" % p['proc_guid'])
                    print("  CreateTime:  %s (GMT)" % self._time_dir_gmt(p['create_time']))
                    print("  ParentPid:   %d" % p['parent'])
                    print("  ParentGuid:  %s" % p['parent_guid'])
                    print("  SID:         %s" % p['sid'].upper())
                    print("  UserName:    %s" % p['username'])
                    print("  ExePath:     %s" % p['path'])
                    print("  CommandLine: %s" % p['command_line'])
                    print("")
                else:
                    print("%5d  %-30s %-20s" % (p['pid'], ntpath.basename(p['path']), p['username']))

        if not opts.verbose:
            print("")

    def do_exec(self, line):
        '''
        Command: exec

        Description:
        Execute a process on the sensor.  This assumes the executable is
        already on the sensor.

        Args:
        exec [OPTS] [process command line and arguments]

        where OPTS are:
         -o <OutputFile> - Redirect standard out and standard error to
              the given file path.
         -d <WorkingDir> - Use the following directory as the process working
              directory
         -w - Wait for the process to complete execution before returning.
        '''

        self._needs_attached()
        #
        # note: option parsing is VERY specific to ensure command args are left
        # as untouched as possible
        #

        OPTS = ['-o', '-d', '-w']
        optOut = None
        optWorkDir = None
        optWait = False

        parts = line.split(' ')
        doParse = True
        while (doParse):
            tok = parts.pop(0)
            if (tok in OPTS):
                if tok == '-w':
                    optWait = True
                if tok == '-o':
                    optOut = parts.pop(0)
                if tok == '-d':
                    optWorkDir = parts.pop(0)
            else:
                doParse = False

        exe = tok

        # ok - now the command (exe) is in tok
        # we need to do some crappy path manipulation
        # to see what we are supposed to execute
        if (self._is_path_absolute(exe)):
            pass
            # do nothin
        elif (self._is_path_drive_relative(exe)):
            # append the current dirve
            exe = self.cwd[:2] + exe
        else:
            # is relative (2 sub-cases)
            ret = self._stat(ntpath.join(self.cwd, exe))
            if (ret == "file"):
                # then a file exist in the current working
                # directory that matches the exe name - execute it
                exe = ntpath.join(self.cwd, exe)
            else:
                # the cwd + exe does not exist - let windows
                # resolve the path
                pass

        # re-format the list and put tok at the front
        if (len(parts) > 0):
            cmdline = exe + ' ' + ' '.join(parts)
        else:
            cmdline = exe
        # print "CMD: %s" % cmdline

        if not optWorkDir:
            optWorkDir = self.cwd

        ret = self.lr_session.create_process(cmdline, wait_for_output=optWait, remote_output_file_name=optOut,
                                             working_directory=optWorkDir)

        if (optWait):
            print(ret)

    def do_get(self, line):
        '''
        Command: get

        Description:
        Get (copy) a file, or parts of file, from the sensor.

        Args:
        get [OPTIONS] <RemotePath> <LocalPath>

        where OPTIONS are:
        -o, --offset : The offset to start getting the file at
        -b, --bytes : How many bytes of the file to get.  The default is all bytes.
        '''
        self._needs_attached()

        p = CliArgs(usage='get [OPTIONS] <RemoteFile> <LocalFile>')
        (opts, args) = p.parse_line(line)

        if len(args) != 2:
            raise CliArgsException("Wrong number of args to get command")

        with open(args[1], "wb") as fout:
            gfile = self._file_path_fixup(args[0])
            shutil.copyfileobj(self.lr_session.get_raw_file(gfile), fout)

    def do_del(self, line):
        '''
        Command: del

        Description:
        Delete a file on the sensor

        Args:
        del <FileToDelete>
        '''

        self._needs_attached()

        if line is None or line == '':
            raise CliArgsException("Must provide argument to del command")

        path = self._file_path_fixup(line)
        self.lr_session.delete_file(path)

    def do_mkdir(self, line):
        '''
        Command: mkdir

        Description:
        Create a directory on the sensor

        Args:
        mdkir <PathToCreate>
        '''

        self._needs_attached()

        if line is None or line == '':
            raise CliArgsException("Must provide argument to mkdir command")

        path = self._file_path_fixup(line)
        self.lr_session.create_directory(path)

    def do_put(self, line):
        '''
        Command: put

        Description
        Put a file onto the sensor

        Args:
        put <LocalFile> <RemotePath>
        '''
        self._needs_attached()

        argv = split_cli(line)

        if len(argv) != 2:
            raise CliArgsException("Wrong number of args to put command (need 2)")

        with open(argv[0], "rb") as fin:
            self.lr_session.put_file(fin, argv[1])

    def _time_dir_gmt(self, unixtime):
        return time.strftime("%m/%d/%Y %I:%M:%S %p", time.gmtime(unixtime))

    def do_dir(self, line):
        self._needs_attached()

        if line is None or line == '':
            line = self.cwd + "\\"

        path = self._file_path_fixup(line)

        if path.endswith('\\'):
            path += '*'

        ret = self.lr_session.list_directory(path)

        print("Directory: %s\n" % path)
        for f in ret:
            timestr = self._time_dir_gmt(f['create_time'])
            if ('DIRECTORY' in f['attributes']):
                x = '<DIR>               '
            else:
                x = '     %15d' % f['size']
            print("%s\t%s %s" % (timestr, x, f['filename']))

        print("")

    def do_kill(self, line):
        '''
        Command: kill

        Description:
        Kill a process on the sensor

        Args:
        kill <Pid>
        '''

        if line is None or line == '':
            raise CliArgsException("Invalid argument passed to kill (%s)" % line)

        self.lr_session.kill_process(line)
    #
    # def do_reg(self, line):
    #     '''
    #     Command: reg
    #
    #     Description:
    #     Perform registry actions (query/add/delete) against
    #     the sensor registry.
    #
    #     Args:
    #     reg [SUB_COMMAND] [SUB_SPECIFIC_OPTIONS]
    #
    #     where SUB_COMMAND is:
    #
    #     add - Add a registry key or value
    #
    #     delete - Delete a registry key or value
    #
    #     query - Query a registry key or value
    #
    #     SUB_SPECIFIC_OPTIONS:
    #
    #     reg add [key] [opts]
    #         -v : add the value instead of the key
    #         -t : value type (REG_DWORD, ....) requires -v
    #         -d : data for the value - converted to appropriate type for
    #              binary data hex encode
    #         -f : force (overwrite value if it already exists)
    #
    #     reg delete [key] [opts]
    #         -v : delete the value instead of the key
    #
    #     reg query [key] [opts]
    #         -v : query a value instead of just a key
    #     '''
    #
    #     self._needs_attached()
    #
    #     args = split_cli(line)
    #     REG_OPTS = ['add', 'delete', 'query']
    #     # pop the first arg off for the sub-command
    #
    #     subcmd = args.pop(0).lower()
    #     if (subcmd not in REG_OPTS):
    #         raise CliArgsException("Invalid reg subcommand! (%s)" % subcmd)
    #
    #     # parse out the args
    #     if (subcmd == 'add'):
    #         p = CliArgs("reg add <Key> [OPTIONS]")
    #         p.add_option('-v', '--value', default=None, help='The value to add (instead of a key)')
    #         p.add_option('-t', '--type', default=None, help='The value type to add')
    #         p.add_option('-d', '--data', default=None, help='The data value to add')
    #         p.add_option('-f', '--force', default=False, action='store_true', help='Overwrite existing value')
    #
    #         (opts, args) = p.parse_args(args)
    #
    #         if (opts.value):
    #             if (not opts.type):
    #                 raise CliArgsException("Must provide a value data type (-t) with -v option")
    #             if (not opts.data):
    #                 raise CliArgsException("Must provide a data value (-d) with -v option")
    #
    #             rval = {}
    #             #rval['value_data'] = opts.data
    #             rval['value_data'] = opts.data
    #             rval['value_type'] = opts.type
    #             rval['overwrite'] = opts.force
    #
    #             path = args[0] + '\\'+ opts.value
    #
    #             self._postCommandAndWait('reg set value', path, args=rval)
    #
    #         else:
    #             # add a key
    #             self._postCommandAndWait('reg create key', args[0])
    #
    #     elif (subcmd == 'delete'):
    #         p = CliArgs("reg delete <Key> [OPTIONS]")
    #         p.add_option('-v', '--value', default=None, help='The value to delete (instead of a key)')
    #         (opts, args) = p.parse_args(args=args)
    #
    #         if (opts.value):
    #             path = args[0] + "\\" + opts.value
    #
    #             print "DeleteValue: %s" % path
    #             self._postCommandAndWait('reg delete value', path)
    #
    #         else:
    #             self._postCommandAndWait('reg delete key', args[0])
    #
    #     else: # query
    #         p = CliArgs('reg query <Key> [OPTIONS]')
    #         p.add_option('-v', '--value', default=None, help='The value to query (instead of a key)')
    #
    #         (opts, args) = p.parse_args(args=args)
    #
    #         if (opts.value):
    #             path = args[0] + "\\" + opts.value
    #             ret = self._postCommandAndWait('reg query value', path)
    #
    #             value = ret.get('value', None)
    #             if (value):
    #                 self._print_reg_value(value)
    #
    #         else:
    #             ret = self._postCommandAndWait('reg enum key', args[0])
    #
    #             subs = ret.get('sub_keys', [])
    #             if (len(subs) > 0):
    #                 print "Keys:"
    #             for k in subs:
    #                 print "%s" % k
    #
    #             values = ret.get('values', [])
    #             if (len(values) > 0):
    #                 print "Values:"
    #             for v in values:
    #                 print "REPR value: %s" + repr(v)
    #                 self._print_reg_value(v)
    #
    # def _print_reg_value(self, value):
    #     type = value['value_type']
    #     name = value['value_name']
    #     rawdata = value['value_data']
    #
    #     # handle default value
    #     if (name == ''):
    #         name = '(Default)'
    #
    #     print "  %-30s %10s %s" %(name, type, rawdata)

    # call the system shell
    def do_shell(self, line):
        '''
        Command: shell

        Description:
        Run a command locally and display the output

        Args:
        shell <Arguments>
        '''
        print(subprocess.Popen(line, shell=True, stdout=subprocess.PIPE).stdout.read())

    # quit handlers
    def do_exit(self, line):
        return self._quit()

    def do_quit(self, line):
        return self._quit()

    def do_EOF(self, line):
        return self._quit()



#!/usr/bin/env python
# Example of using cbapi to get event data
# usage:
# python list_events.py --hostname <hostname> --start<YYYY-MM-DD> --end<YYYY-MM-DD>


# Notes on this script:
#  - script can only pull up to 2 weeks of events at one time ( this is an API limitation)
#  - if no data exists between the start and end range, the script will pull no data.
#  - script can be run with no arguments, it will return events from all endpoints for the past 2 weeks

import sys
import re
from datetime import datetime
from cbapi.psc.defense.models import Event
from cbapi.example_helpers import build_cli_parser, get_cb_defense_object


# Function to format epoch time
def convert_time(epoch_time):
    converted_time = datetime.fromtimestamp(int(epoch_time / 1000.0)).strftime(' %b %d %Y %H:%M:%S')
    return converted_time


# Function to strip HTML from a string.
def strip_html(string):
    p = re.compile(r'<.*?>')
    return p.sub('', string)


def main():
    parser = build_cli_parser("List Events for a device")
    event_options = parser.add_mutually_exclusive_group(required=False)
    event_date_options = parser.add_argument_group("Date Range Arguments")
    event_date_options.add_argument("--start", help="start time")
    event_date_options.add_argument("--end", help="end time")
    event_options.add_argument("-n", "--hostname", help="Hostname")

    args = parser.parse_args()
    cb = get_cb_defense_object(args)

    if args.hostname:
        events = list(cb.select(Event).where("hostNameExact:{0}".format(args.hostname)))
    elif args.start and args.end:
        # flipped the start and end arguments around so script can be called with the start date being
        # the earliest date. it's just easier on the eyes for most folks.

        events = list(cb.select(Event).where("startTime:{0}".format(args.end))) and (
            cb.select(Event).where("endTime:{0}".format(args.start)))
    else:
        events = list(cb.select(Event))

    for event in events:
        # convert event and create times
        event_time = str(convert_time(event.createTime))
        create_time = str(convert_time(event.eventTime))

        # stripping HTML tags out of the long description
        long_description = strip_html(event.longDescription)

        # format and print out the event time, Event ID, Creation time, Event type and Description
        print("{0:^25}{1:^25}{2:^32}{3}".format("Event Time", "Event ID", "Create Time", "Event Type"))
        print("{0} | {1} | {2} | {3}".format(event_time, event.eventId, create_time, event.eventType))
        print("{0:50}".format("                                    "))
        print("{0} {1}".format("Description: ", long_description))
        print("{0:50}".format("------------------------------------"))
        print("{0:50}".format("                                    "))


if __name__ == "__main__":
    sys.exit(main())



def setup_parser_with_basic_criteria(parser):
    parser.add_argument("-q", "--query", help="Query string for looking for alerts")
    parser.add_argument("--category", action="append", choices=["THREAT", "MONITORED", "INFO",
                                                                "MINOR", "SERIOUS", "CRITICAL"],
                        help="Restrict search to the specified categories")
    parser.add_argument("--deviceid", action="append", type=int, help="Restrict search to the specified device IDs")
    parser.add_argument("--devicename", action="append", type=str, help="Restrict search to the specified device names")
    parser.add_argument("--os", action="append", choices=["WINDOWS", "ANDROID", "MAC", "IOS", "LINUX", "OTHER"],
                        help="Restrict search to the specified device operating systems")
    parser.add_argument("--osversion", action="append", type=str,
                        help="Restrict search to the specified device operating system versions")
    parser.add_argument("--username", action="append", type=str, help="Restrict search to the specified user names")
    parser.add_argument("--group", action="store_true", help="Group results")
    parser.add_argument("--alertid", action="append", type=str, help="Restrict search to the specified alert IDs")
    parser.add_argument("--legacyalertid", action="append", type=str,
                        help="Restrict search to the specified legacy alert IDs")
    parser.add_argument("--severity", type=int, help="Restrict search to the specified minimum severity level")
    parser.add_argument("--policyid", action="append", type=int, help="Restrict search to the specified policy IDs")
    parser.add_argument("--policyname", action="append", type=str, help="Restrict search to the specified policy names")
    parser.add_argument("--processname", action="append", type=str,
                        help="Restrict search to the specified process names")
    parser.add_argument("--processhash", action="append", type=str,
                        help="Restrict search to the specified process SHA-256 hash values")
    parser.add_argument("--reputation", action="append", choices=["KNOWN_MALWARE", "SUSPECT_MALWARE", "PUP",
                                                                  "NOT_LISTED", "ADAPTIVE_WHITE_LIST",
                                                                  "COMMON_WHITE_LIST", "TRUSTED_WHITE_LIST",
                                                                  "COMPANY_BLACK_LIST"],
                        help="Restrict search to the specified reputation values")
    parser.add_argument("--tag", action="append", type=str, help="Restrict search to the specified tag values")
    parser.add_argument("--priority", action="append", choices=["LOW", "MEDIUM", "HIGH", "MISSION_CRITICAL"],
                        help="Restrict search to the specified priority values")
    parser.add_argument("--threatid", action="append", type=str, help="Restrict search to the specified threat IDs")
    parser.add_argument("--type", action="append", choices=["CB_ANALYTICS", "VMWARE", "WATCHLIST"],
                        help="Restrict search to the specified alert types")
    parser.add_argument("--workflow", action="append", choices=["OPEN", "DISMISSED"],
                        help="Restrict search to the specified workflow statuses")


def setup_parser_with_cbanalytics_criteria(parser):
    setup_parser_with_basic_criteria(parser)
    parser.add_argument("--blockedthreat", action="append", choices=["UNKNOWN", "NON_MALWARE", "NEW_MALWARE",
                                                                     "KNOWN_MALWARE", "RISKY_PROGRAM"],
                        help="Restrict search to the specified threat categories that were blocked")
    parser.add_argument("--location", action="append", choices=["ONSITE", "OFFSITE", "UNKNOWN"],
                        help="Restrict search to the specified device locations")
    parser.add_argument("--killchain", action="append", choices=["RECONNAISSANCE", "WEAPONIZE", "DELIVER_EXPLOIT",
                                                                 "INSTALL_RUN", "COMMAND_AND_CONTROL", "EXECUTE_GOAL",
                                                                 "BREACH"],
                        help="Restrict search to the specified kill chain status values")
    parser.add_argument("--notblockedthreat", action="append", choices=["UNKNOWN", "NON_MALWARE", "NEW_MALWARE",
                                                                        "KNOWN_MALWARE", "RISKY_PROGRAM"],
                        help="Restrict search to the specified threat categories that were NOT blocked")
    parser.add_argument("--policyapplied", action="append", choices=["APPLIED", "NOT_APPLIED"],
                        help="Restrict search to the specified policy-application status values")
    parser.add_argument("--reason", action="append", type=str, help="Restrict search to the specified reason codes")
    parser.add_argument("--runstate", action="append", choices=["DID_NOT_RUN", "RAN", "UNKNOWN"],
                        help="Restrict search to the specified run states")
    parser.add_argument("--sensoraction", action="append", choices=["POLICY_NOT_APPLIED", "ALLOW", "ALLOW_AND_LOG",
                                                                    "TERMINATE", "DENY"],
                        help="Restrict search to the specified sensor actions")
    parser.add_argument("--vector", action="append", choices=["EMAIL", "WEB", "GENERIC_SERVER", "GENERIC_CLIENT",
                                                              "REMOTE_DRIVE", "REMOVABLE_MEDIA", "UNKNOWN",
                                                              "APP_STORE", "THIRD_PARTY"],
                        help="Restrict search to the specified threat cause vectors")


def setup_parser_with_vmware_criteria(parser):
    setup_parser_with_basic_criteria(parser)
    parser.add_argument("--groupid", action="append", type=int,
                        help="Restrict search to the specified AppDefense alarm group IDs")


def setup_parser_with_watchlist_criteria(parser):
    setup_parser_with_basic_criteria(parser)
    parser.add_argument("--watchlistid", action="append", type=str,
                        help="Restrict search to the specified watchlists by ID")
    parser.add_argument("--watchlistname", action="append", type=str,
                        help="Restrict search to the specified watchlists by name")


def load_basic_criteria(query, args):
    if args.query:
        query = query.where(args.query)
    if args.category:
        query = query.set_categories(args.category)
    if args.deviceid:
        query = query.set_device_ids(args.deviceid)
    if args.devicename:
        query = query.set_device_names(args.devicename)
    if args.os:
        query = query.set_device_os(args.os)
    if args.osversion:
        query = query.set_device_os_versions(args.osversion)
    if args.username:
        query = query.set_device_username(args.username)
    if args.group:
        query = query.set_group_results(True)
    if args.alertid:
        query = query.set_alert_ids(args.alertid)
    if args.legacyalertid:
        query = query.set_legacy_alert_ids(args.legacyalertid)
    if args.severity:
        query = query.set_minimum_severity(args.severity)
    if args.policyid:
        query = query.set_policy_ids(args.policyid)
    if args.policyname:
        query = query.set_policy_names(args.policyname)
    if args.processname:
        query = query.set_process_names(args.processname)
    if args.processhash:
        query = query.set_process_sha256(args.processhash)
    if args.reputation:
        query = query.set_reputations(args.reputation)
    if args.tag:
        query = query.set_tags(args.tag)
    if args.priority:
        query = query.set_target_priorities(args.priority)
    if args.threatid:
        query = query.set_threat_ids(args.threatid)
    if args.type:
        query = query.set_types(args.type)
    if args.workflow:
        query = query.set_workflows(args.workflow)


def load_cbanalytics_criteria(query, args):
    load_basic_criteria(query, args)
    if args.blockedthreat:
        query = query.set_blocked_threat_categories(args.blockedthreat)
    if args.location:
        query = query.set_device_locations(args.location)
    if args.killchain:
        query = query.set_kill_chain_statuses(args.killchain)
    if args.notblockedthreat:
        query = query.set_not_blocked_threat_categories(args.notblockedthreat)
    if args.policyapplied:
        query = query.set_policy_applied(args.policyapplied)
    if args.reason:
        query = query.set_reason_code(args.reason)
    if args.runstate:
        query = query.set_run_states(args.runstate)
    if args.sensoraction:
        query = query.set_sensor_actions(args.sensoraction)
    if args.vector:
        query = query.set_threat_cause_vectors(args.vector)


def load_vmware_criteria(query, args):
    load_basic_criteria(query, args)
    if args.groupid:
        query = query.set_group_ids(args.groupid)


def load_watchlist_criteria(query, args):
    load_basic_criteria(query, args)
    if args.watchlistid:
        query = query.set_watchlist_ids(args.watchlistid)
    if args.watchlistname:
        query = query.set_watchlist_names(args.watchlistname)



#!/usr/bin/env python

import sys
from cbapi.response.models import Binary
from cbapi.example_helpers import build_cli_parser, get_cb_response_object
import shutil


def main():
    parser = build_cli_parser()
    parser.add_argument("--md5", help="binary query", required=True)
    parser.add_argument("--filename", help="local filename to save the binary as", required=True)
    args = parser.parse_args()

    cb = get_cb_response_object(args)
    binary = cb.select(Binary, args.md5)
    shutil.copyfileobj(binary.file, open(args.filename, "wb"))

    print("-> Downloaded binary %s [%u bytes]" % (args.md5, binary.size))

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import threading
import Queue
import logging
from collections import defaultdict
from functools import wraps
import time
from copy import deepcopy
import traceback

log = logging.getLogger(__name__)


class EventRegistry(threading.Thread):
    def __init__(self):
        super(EventRegistry, self).__init__()

        self._registry = None
        self._registry_lock = threading.RLock()
        self._error_lock = threading.RLock()
        self._errors = []
        self.clear()
        self._callback_queue = Queue.Queue()
        self.daemon = True

    def register(self, event_type, func, *args, **kwargs):
        with self._registry_lock:
            log.debug("Registering function {0} for event_type {1}".format(func, event_type))
            self._registry[event_type].append({"func": func, "args": args, "kwargs": kwargs})

    @property
    def event_types(self):
        with self._registry_lock:
            return self._registry.keys()

    def eval_callback(self, event_type, event_data, cb):
        with self._registry_lock:
            callbacks = self._registry.get(event_type, [])

        for callback in callbacks:
            self._callback_queue.put((callback, cb, event_type, event_data))

    def clear(self):
        with self._registry_lock:
            self._registry = defaultdict(list)

    # TODO: evaluate concurrent.futures.ThreadPoolExecutor for this
    def run(self):
        log.debug("starting event registry thread")
        while True:
            callback, cb, event_type, event_data = self._callback_queue.get()

            kwargs = callback["kwargs"]
            kwargs["cb"] = cb
            kwargs["event_type"] = event_type
            kwargs["event_data"] = event_data

            try:
                callback["func"](*callback["args"], **kwargs)
            except Exception:
                with self._error_lock:
                    self._errors.append({"exception": traceback.format_exc(), "timestamp": time.time(),
                                         "callback_func": callback["func"].__name__,
                                         "event_type": event_type, "event_data": event_data})

    @property
    def errors(self):
        with self._error_lock:
            errs = deepcopy(self._errors)
        return errs


registry = EventRegistry()


def on_event(event_type=None):
    def decorator(func):
        registry.register(event_type, func)

        @wraps(func)
        def f(*args, **kwargs):
            return func(*args, **kwargs)

        return f
    return decorator


registry.start()



from ..utils import convert_query_params
from ..query import PaginatedQuery
from cbapi.six.moves import urllib
from distutils.version import LooseVersion
from ..errors import ApiError
import copy

import logging

log = logging.getLogger(__name__)


class Query(PaginatedQuery):
    """Represents a prepared query to the Carbon Black Enterprise Response server.

    This object is returned as part of a :py:meth:`CbEnterpriseResponseAPI.select`
    operation on Process and Binary objects from the Carbon Black
    Enterprise Response server. You should not have to create this class yourself.

    The query is not executed on the server until it's accessed, either as an iterator (where it will generate values
    on demand as they're requested) or as a list (where it will retrieve the entire result set and save to a list).
    You can also call the Python built-in ``len()`` on this object to retrieve the total number of items matching
    the query.

    The syntax for query :py:meth:where and :py:meth:sort methods can be found in the
    `Query Reference <http://developer.carbonblack.com/resources/query_overview.pdf>`_ posted on the Carbon Black
    Developer Network website.

    Examples::

    >>> cb = CbEnterpriseResponseAPI()
    >>> query = cb.select(Process)                      # returns a Query object matching all Processes
    >>> query = query.where("process_name:notepad.exe") # add a filter to this Query
    >>> query = query.sort("last_update desc")          # sort by last update time, most recent first
    >>> for proc in query:                              # uses the iterator to retrieve all results
    >>>     print("{0} {1}".format(proc.username, proc.hostname))
    >>> processes = query[:10]                          # retrieve the first ten results
    >>> len(query)                                      # retrieve the total count

    Notes:
        - The slicing operator only supports start and end parameters, but not step. ``[1:-1]`` is legal, but
          ``[1:2:-1]`` is not.
        - You can chain where clauses together to create AND queries; only objects that match all ``where`` clauses
          will be returned.
    """
    def __init__(self, doc_class, cb, query=None, raw_query=None):
        super(Query, self).__init__(doc_class, cb, query=query)

        if raw_query:
            self._raw_query = urllib.parse.parse_qs(raw_query)
        else:
            self._raw_query = None

        self._sort_by = getattr(self._doc_class, 'default_sort', None)
        self._default_args = {"cb.urlver": 1}

        # FIX: Cb Response server version 5.1.0-3 throws an exception after returning HTTP 504 when facet=false in the
        # HTTP request. Work around this by only setting facet=false on 5.1.1 and above server versions.
        if self._cb.cb_server_version >= LooseVersion('5.1.1'):
            self._default_args["facet"] = "false"

    def _clone(self):
        nq = self.__class__(self._doc_class, self._cb)
        nq._query = self._query
        nq._raw_query = None
        if self._raw_query:
            nq._raw_query = self._raw_query

        nq._sort_by = self._sort_by
        nq._default_args = copy.deepcopy(self._default_args)
        nq._batch_size = self._batch_size
        return nq

    def sort(self, new_sort):
        """Set the sort order for this query.

        :param str new_sort: New sort order - see the
            `Query Reference <http://developer.carbonblack.com/resources/query_overview.pdf>`_.
        :return: Query object
        :rtype: :py:class:`Query`
        """
        new_sort = new_sort.strip()

        nq = self._clone()
        if len(new_sort) == 0:
            nq._sort_by = None
        else:
            nq._sort_by = new_sort
        return nq

    @property
    def webui_link(self):
        return "{0:s}/#/search?{1}".format(self._cb.url, urllib.parse.urlencode(
            convert_query_params(self._get_query_parameters())))

    def and_(self, new_query):
        """Add a filter to this query. Equivalent to calling :py:meth:`where` on this object.

        :param str new_query: Query string - see the
            `Query Reference <http://developer.carbonblack.com/resources/query_overview.pdf>`_.
        :return: Query object
        :rtype: :py:class:`Query`
        """
        return self.where(new_query)

    def where(self, new_query):
        """Add a filter to this query.

        :param str new_query: Query string - see the
            `Query Reference <http://developer.carbonblack.com/resources/query_overview.pdf>`_.
        :return: Query object
        :rtype: :py:class:`Query`
        """
        if self._raw_query:
            raise ApiError("Cannot call .where() on a raw query")

        nq = self._clone()
        if nq._query and len(nq._query) > 0:
            nq._query = "{0:s} {1:s}".format(self._query, new_query)
        else:
            nq._query = new_query

        return nq

    def facets(self, *args):
        """Retrieve a dictionary with the facets for this query.

        :param args: Any number of fields to use as facets
        :return: Facet data
        :rtype: dict
        """
        # TODO: make this interface better
        qargs = self._default_args.copy()
        qargs['facet'] = 'true'
        qargs['start'] = 0
        qargs['rows'] = 0
        qargs['facet.field'] = list(args)

        if self._query:
            qargs['q'] = self._query

        query_params = convert_query_params(qargs)
        return self._cb.get_object(self._doc_class.urlobject, query_parameters=query_params).get('facets', {})

    def _get_query_parameters(self):
        if self._raw_query:
            args = self._raw_query.copy()
        else:
            args = self._default_args.copy()
            if self._query:
                args['q'] = self._query
            else:
                args['q'] = ''

        return args

    def _count(self):
        if self._count_valid:
            return self._total_results

        args = self._get_query_parameters()
        args['start'] = 0
        args['rows'] = 0

        qargs = convert_query_params(args)

        self._total_results = self._cb.get_object(self._doc_class.urlobject,
                                                  query_parameters=qargs).get('total_results', 0)

        self._count_valid = True
        return self._total_results

    def _search(self, start=0, rows=0):
        # iterate over total result set, 100 at a time

        args = self._get_query_parameters()
        args['start'] = start

        if self._sort_by:
            args['sort'] = self._sort_by

        if rows:
            args['rows'] = min(rows, self._batch_size)
        else:
            args['rows'] = self._batch_size

        still_querying = True
        current = start
        numrows = 0

        while still_querying:
            qargs = convert_query_params(args)
            result = self._cb.get_object(self._doc_class.urlobject, query_parameters=qargs)

            self._total_results = result.get('total_results')
            self._count_valid = True

            results = result.get('results', [])

            for item in results:
                yield item
                current += 1
                numrows += 1
                if rows and numrows == rows:
                    still_querying = False
                    break

            args['start'] = current

            if current >= self._total_results:
                break
            if not results:
                log.debug("server reported total_results overestimated the number of results for this query by {0}"
                          .format(self._total_results - current))
                log.debug("resetting total_results for this query to {0}".format(current))
                self._total_results = current
                break



#!/usr/bin/env python

import sys
from time import sleep
from cbapi.example_helpers import build_cli_parser, get_cb_psc_object
from cbapi.psc.models import WorkflowStatus


def main():
    parser = build_cli_parser("Bulk update the status of alerts by threat ID")
    parser.add_argument("-T", "--threatid", action="append", type=str, required=True,
                        help="Threat IDs to update the alerts for")
    parser.add_argument("-R", "--remediation", help="Remediation message to store for the selected alerts")
    parser.add_argument("-C", "--comment", help="Comment message to store for the selected alerts")
    operation = parser.add_mutually_exclusive_group(required=True)
    operation.add_argument("--dismiss", action="store_true", help="Dismiss all selected alerts")
    operation.add_argument("--undismiss", action="store_true", help="Undismiss all selected alerts")

    args = parser.parse_args()
    cb = get_cb_psc_object(args)

    if args.dismiss:
        reqid = cb.bulk_threat_dismiss(args.threatid, args.remediation, args.comment)
    elif args.undismiss:
        reqid = cb.bulk_threat_update(args.threatid, args.remediation, args.comment)
    else:
        raise NotImplementedError("one of --dismiss or --undismiss must be specified")

    print("Submitted query with ID {0}".format(reqid))
    statobj = cb.select(WorkflowStatus, reqid)
    while not statobj.finished:
        print("Waiting...")
        sleep(1)
    if statobj.errors:
        print("Errors encountered:")
        for err in statobj.errors:
            print("\t{0}".format(err))
    if statobj.failed_ids:
        print("Failed alert IDs:")
        for i in statobj.failed_ids:
            print("\t{0}".format(err))
    print("{0} total alert(s) found, of which {1} were successfully changed"
          .format(statobj.num_hits, statobj.num_success))


if __name__ == "__main__":
    sys.exit(main())
